#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#define ASIZE 100


void bucketsort(int *a) {
	/* Pedimos los 4 kilobytes
	 * de memoria como un arreglo
	 * de chars (1 byte cada uno)
	 */
	char *mem = (char *)malloc(4096);
	
	int i,j;
	
	/* Apagamos todos los bits de la
	 * memoria pedida. Esta parte
	 * toma tiempo constante.
	 */
	i = 0;
	for(;;) {
		mem[i++] = 0;

		/* Preguntamos si el bit en la
		 * posicin 12 (numerando los
		 * bits desde 0 y de derecha a
		 * izquierda)
		 * est encendido. El primer
		 * nmero con el bit 12 encendido
		 * es el 4096.
		 *
		 * 4096 es en binario,
		 * 1000000000000
		 * ^ <- Posicin 12.
		 *
		 * Por lo tanto cuando el nmero
		 * sea 4096 el ciclo parar.
		 */
		if(i&(1<<12))
			break;
	}
	
	/* Vamos a recorrer el arreglo
	 * Si el elemento es a[i], entonces
	 * encenderemos el bit a[i]-simo
	 * de la memoria pedida
	 */
	
	/* Esta parte toma tiempo proporcional
	 * al nmero de elementos del arreglo
	 */
	i = 0;
	for(;;) {
		/* El ltimo elemento del arreglo es
		 * -1. Por lo tanto el ciclo debe terminar
		 *  cuando este valor sea visto en el elemento
		 *  actual del arreglo, a[i].
		 *
		 *  -1 en binario es ~1 + 1
		 *  -1	= ~1 + 1
		 *  	= ~(00000001) + 1
		 *  	=   11111110  + 1
		 *  	=   11111111
		 *
		 *  (En cualquier largo de palabra el -1 siempre va a ser
		 *  una secuencia de slo 1s)
		 *
		 *  Pero podemos hacer el -1 ms fcil con ~0, que tambin
		 *  es una tira de bits todos encendidos.
		 *
		 *  Ahora veamos que ocurre cuando aplicamos el OR exclusivo
		 *  de bits a ~0
		 *     	
		 *     11111111 (~0)
		 *     xxxxxxxx (a[i])
		 *   ^ --------
		 *     ????????
		 *
		 * Donde las x son los bits de a[i]. Qu pasa si a[i] es -1?
		 *
		 *     11111111 (~0)
		 *     11111111 (-1 = a[i])
		 *   ^ --------
		 *     00000000
		 * 
		 * El resultado es una tira de bits apagados. El nico nmero que en la
		 * operacin anterior puede generar una tira de bits apagados al ser
		 * operado con un OR exclusivo con ~0 es precisamente el -1.
		 *
		 * Cualquier otro nmero no genera una tira de bits apagados. Por ejemplo
		 * cualquier nmero distinto de -1, tiene en algn bit un cero (pues si
		 * los tiene todos en 1, es el -1 :P ). Si tiene algn bit en cero, lo que
		 * ocurre es,
		 *
		 *     11111111 (~0)
		 *     xxxxx0xx (a[i] != -1)
		 *   ^ --------
		 *     ?????1??
		 *
		 * Y no resulta una tira de bits apagados (hay al menos uno encendido).
		 * 
		 * En el caso que a[i] sea -1 entonces el resultado de la operacin
		 * a[i]^(~0) es 0. Si 0 es falso en C, entonces negando lo anterior
		 * con !, obtenemos la condicin trmino para el ciclo.
		 */
		
		if(!(a[i]^(~0))) /* Si (a[i] == -1) el arreglo se acab*/
			break;
		
		/* Ahora veamos un elemento que debe
		 * ser ordenado, a[i] != -1.
		 *
		 * Encenderemos el bit a[i]-simo.
		 * Este bit se encuentra en el byte a[i]/8 = a[i] >> 3
		 * del arreglo, y su posicin dentro del byte es
		 * (a[i]%8) de derecha a izquierda (numerados desde
		 * 0 a 7).
		 * Para encender el bit (a[i]%8) debemos hacer
		 * un OR de bits, |, con la mscara que tenga
		 * slo ese bit encendido.
		 * 
		 * La mscara que tiene slo el bit (a[i]%8)
		 * encendido es (1 << (a[i]%8)).
		 * 
		 * Teniendo la mscara y el byte determinados
		 * realizamos,
		 *
		 * byte = byte | mascara;
		 *        ^^^^^^^^^^^^^^
		 *        byte con un bit
		 *        encendido gracias
		 *        a la mscara
		 *
		 * Lo que se abrevia como,
		 *
		 * byte |= mascara;
		 *
		 * Para calcular (a[i]%8) debemos fijarnos que
		 * la expresin es equivalente a ver los primeros
		 * 3 bits de (a[i]%8), pues siempre sern el resto
		 * de dividir por 8 (pues con 3 bits puedo escribir
		 * desde el 0 al 7).
		 *
		 * Una analoga de lo anterior es que en base 10,
		 * al ver el resto de dividir por una potencia de
		 * la base (10 a la algo), lo que importa son los
		 * dgitos menos significativos hasta
		 * el dgito necesario para escribir la potencia
		 * de 10 por la que estamos dividiendo, por ejemplo
		 *
		 * 89398 % 100 = 00098 
		 *
		 * Pues me interesa el trozo hasta el segundo dgito, pues
		 * con el tercer dgito ya puedo escribir 100, que es por
		 * lo que estoy dividiendo.
		 *
		 * De la misma manera, en binario, el 8 se escribe como
		 * 1000, entonces me interesarn los bits hasta antes
		 * del bit que me sirve para escribir el 8, pues eso
		 * ser siempre lo que sobre de dividir por 8.
		 *
		 * Necesitamos entonces la mscara 00000111, para
		 * que al hacer a[i] & 00000111 queden prendidos
		 * solo los 3 primeros bits de a[i].
		 *
		 * La mscara se construye as,
		 * ~(~0 << 3)	= ~(11111111 << 3)
		 * 		= ~(11111000)
		 * 		=   00000111
		 *
		 * Luego, a[i]%8 = (a[i] & ~(~0 << 3))
		 * 
		 */ 
		
		mem[a[i] >> 3] |= (1 << (a[i] & ~(~0 << 3)));
		++i;
	}
	

	/* Esta parte consiste en recorrer la memoria 
	 * EN ORDEN y mirar todos los bits que estn prendidos
	 * y asignarlos de vuelta al arreglo hasta que este se
	 * llene.
	 *
	 * Si pensamos que el nmero ms grande que puede salir
	 * es 32767, y vamos a verificar desde 0 hasta el nmero
	 * ms grande, en el peor de los casos el ciclo toma
	 * tiempo constante (32769 iteraciones [32768 elementos + 1 por
	 * la marca de final del arreglo]).
	 */
	i = 0;
	j = 0;
	for(;;) {
		
		if(!(a[i]^(~0)))
			break;
		
		for(;;) {

			/* En el byte j/8 verificamos si el bit j%8 est prendido,
			 * es decir vemos si el bit que le corresponde a j
			 * est prendido*/

			if(mem[j >> 3] & (1 << (j & ~(~0 << 3)))) { 

				/* Si el bit est prendido, entonces
				 * el elemento est en el arreglo*/

				a[i++] = j++;
				break;
			}

			/* Si el bit no est prendido sigo buscando uno que s
			 * lo est*/

			++j;
		}
	}
	/* Alguien dijo valgrind? xD*/
	free(mem);
}

int main() {
	/*Probar las funciones*/
	int ordenar[ASIZE];
	int i = 0;
	
	
	/* Genero nmeros al lote*/
	srand(time(NULL));
	for(i = 0; i < ASIZE; ++i) {
		ordenar[i] = (16383 + (i*17*rand()%16383)*(i%2?1:-1));
	}
	
	/* Marco el final de arreglo*/
	ordenar[ASIZE - 1] = -1;
	
	/* Imprimo el orden inicial*/
	for(i = 0; i < ASIZE; ++i)
		printf("%d\n", ordenar[i]);
	
	/* Separador visual*/
	printf("#####\n");
	
	/* Ordeno*/
	bucketsort(ordenar);
	
	/* Muestro los elementos ordenados*/
	for(i = 0; i < ASIZE; ++i)
		printf("%d\n", ordenar[i]);
	

	/* :P */
	return EXIT_SUCCESS;
}
